use crate::deref_as;
use bevy::math::Vec3Swizzles;
use bevy::prelude::{Changed, Color, Component, Entity, Query, Rect, Transform, Vec2};

#[derive(Clone, Copy, Debug, Component)]
pub struct BoxSize(Vec2);
deref_as!(BoxSize => Vec2);
impl From<Vec2> for BoxSize {
	fn from(value: Vec2) -> Self {
		BoxSize(value)
	}
}

pub fn check_box_collisions(first: Rect, second: Rect) -> bool {
	first.min.x < second.max.x
		&& first.max.x > second.min.x
		&& first.min.y < second.max.y
		&& first.max.y > second.min.y
}

#[derive(Clone, Copy, Debug, Component)]
pub enum CollisionGroup {
	Player,
	Pickup,
	FriendlyProjectile,
	HostileProjectile,
	Enemy,
}

impl CollisionGroup {
	pub fn color(&self) -> Color {
		use CollisionGroup::*;
		match self {
			Player => Color::rgb_u8(0, 116, 217),
			Pickup => Color::rgb_u8(177, 13, 201),
			FriendlyProjectile => Color::rgb_u8(46, 204, 64),
			HostileProjectile => Color::rgb_u8(255, 65, 54),
			Enemy => Color::rgb_u8(133, 20, 75),
		}
	}

	pub fn interacts_with(&self, other: CollisionGroup) -> bool {
		use CollisionGroup::*;

		matches!(
			(*self, other),
			(FriendlyProjectile, Enemy)
				| (Enemy, FriendlyProjectile)
				| (Player, Pickup)
				| (Pickup, Player)
				| (HostileProjectile, Player)
				| (Player, HostileProjectile)
		)
	}
}

pub fn check_entity_collisions(
	moved_colliders: Query<(Entity, &Transform, &BoxSize, &CollisionGroup), Changed<Transform>>,
	all_colliders: Query<(Entity, &Transform, &BoxSize, &CollisionGroup)>,
) {
	for (moved_entity, moved_transform, moved_size, moved_group) in &moved_colliders {
		for (target_entity, target_transform, target_size, target_group) in &all_colliders {
			if moved_entity == target_entity {
				continue;
			}

			let moved_box = Rect::from_center_size(moved_transform.translation.xy(), **moved_size);
			let target_box =
				Rect::from_center_size(target_transform.translation.xy(), **target_size);

			if moved_group.interacts_with(*target_group)
				&& check_box_collisions(moved_box, target_box)
			{
				log::info!(
					"Two colliding entities! {:?} and {:?}",
					target_entity,
					moved_entity
				);
			}
		}
	}
}
